// CompoundComponent.cpp: implementation of the CompoundComponent class.
//
//////////////////////////////////////////////////////////////////////

#include "CompoundComponent.h"
#include "BasicComponent.h"
#include "ShiftedComponent.h"
#include "ComponentRegister.h"
#include "ComponentVariant.h"
#include "Writer.h"
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CompoundComponent::CompoundComponent(long numComponents, long maxVariants, long matrixSize)
{
	m_area = -1;
	m_maxComponents = numComponents;
	m_maxVariants = maxVariants;
	m_list = new ShiftedComponent[m_maxComponents];
	assert(m_list != NULL);

	for (long i = 0; i < m_maxComponents; i++)
	{
		m_list[i].setVariants(maxVariants);
	}

	if (matrixSize > 0)
	{
		m_matrixSize = matrixSize;
		m_variantMatrix = new bool[m_matrixSize * m_matrixSize];
		assert(m_variantMatrix != NULL);
		m_variantVector = new VariantRecord[m_matrixSize];
		assert(m_variantVector != NULL);
		m_componentVector = new ComponentRecord[m_maxComponents];
		assert(m_componentVector != NULL);
	}
	else
	{
		m_matrixSize = 0;
		m_variantMatrix = NULL;
		m_variantVector = NULL;
		m_componentVector = NULL;
	}
	reset();
}

CompoundComponent::~CompoundComponent()
{
	delete [] m_list;
	delete [] m_variantMatrix;
	delete [] m_variantVector;
	delete [] m_componentVector;
}

void CompoundComponent::print(long setNo)
{
	char buf[80];
	if (setNo != 0)
	{
		sprintf(buf,"(%d @ %d)", getArea(), setNo);
	}
	else
	{
		sprintf(buf,"(%d)", getArea());
	}
	Writer::print(buf);

	for (long i = 0; i < m_numComponents; i++)
	{
		Writer::print(" ");
		m_list[i].print();
	}
}

bool CompoundComponent::getCell(long x, long y)
{
	for (long i = 0; i < m_numComponents; i++)
	{
		if (m_list[i].getCell(x, y))
		{
			return true;
		}
	}
	return false;
}

CompoundComponent * CompoundComponent::clone(Direction dir)
{
	CompoundComponent * cc = new CompoundComponent(m_maxComponents, m_maxVariants, 0);
	assert(cc != NULL);

	if (m_numComponents > 0)
	{
		cc->addVariant(0, 0, dir, m_list[0].getBasicComponent(), m_list[0].getVariant());
	}
	for (long i = 1; i < m_numComponents; i++)
	{
		cc->addVariant(
			m_list[i].getBasicComponent(),
			m_list[i].getVariant());
	}
	return cc;
}

// this function checks if components can be put this way
// returns zero if yes
// otherwise it returns number of components that
// must be removed from the end

// this method may seem ugly due to its usage of labels and goto statements
// but keep in mind that the algorithm spends over 90% of time here
// so I tried to do it efficient rather than nice

//#define CHECKDEBUG

long CompoundComponent::check(bool optimize)
{
	if (m_numComponents == 0)
	{
		// let's declare empty list OK
		return 0;
	}

	if (m_matrixSize == 0)
	{
		// don't do anything if the memory is not prepared
		return m_numComponents;
	}

#ifdef CHECKDEBUG
	Writer::print("------------- Starting Check() --------------\n");
#endif
	bool closed = (m_list[0].getSx() == m_curX)
		&& (m_list[0].getSy() == m_curY)
		&& (m_list[0].getSDirection() == m_curDir);

	long maxDepth = 0;
	long minArea = -1;

	long stackDepth = 0;
	long numVariants = 0;
	VariantRecord * lastVariant = m_variantVector;
	ComponentRecord * stackTop = m_componentVector;
	ComponentRecord * realTop; // for optimization run
	ComponentRecord * lastComponent = m_componentVector;

	// calculate minimum result size
	long xMin, xMax, yMin, yMax;
	getMinSize(xMin, xMax, yMin, yMax);


NextComponent:
	// invariant:
	// components under stack level are working together
	// there is nothing active on stack level

	// first check if we are not done

	if (stackDepth == m_numComponents)
	{
		// we have a solution!
#ifdef CHECKDEBUG		
		{
			Writer::print("Solution found\n");
		}
#endif
		// move the solution to m_list
		ComponentRecord * cx = m_componentVector;
		long c = m_numComponents;
		do
		{
			m_list[cx->number].setVariantNo(cx->activeVariant->number);
			++cx;
		} while (cx != stackTop);

		if (optimize)
		{
			// solution exists
			// but we need to find the smallest one
			// and this requires slightly different approach
			goto StartOptimize;
		}

		// if we don't need to optimize, we can just exit

		return 0;
	}

	// solution not reached

	// if we are increasing the component count, some actions have to be done, too
	// if not, then everything is ready
	if (stackTop != lastComponent)
	{
		// just initialize the active variant
		stackTop->activeVariant = stackTop->lastVariant;
	}
	else
	{
		// let's add another component
		++lastComponent;

		if (stackDepth == 0)
		{
			// it's the first component - let's just add its variants
			m_list[0].loadVariants(NULL, NULL);
		}
		else
		{
			// take just the previous component into account
			m_list[stackDepth].loadVariants(m_list[stackDepth - 1].getBasicComponent(), NULL);
		}

		stackTop->number = stackDepth;

		// variants loaded - get their number
		stackTop->numVariants = m_list[stackDepth].getNumVariants();

		if (stackTop->numVariants == 0)
		{
			// no variant available - stop
			return m_numComponents - maxDepth;
		}

		// at first, all variants are active and free
		stackTop->activeVariants = stackTop->numVariants;

		// and prepare the variant number for the next component
		stackTop->firstVariant = lastVariant;
		lastVariant += stackTop->numVariants;
		numVariants += stackTop->numVariants;
		stackTop->lastVariant = lastVariant;
		stackTop->activeVariant = lastVariant;
		stackTop->reorder = stackTop; // reorder points at itself first

		// go through all new variants
		{
			long v = stackTop->numVariants;
			VariantRecord * vr = stackTop->firstVariant;
			do
			{
				// Variants must be planted in opposite order
				// because the search also takes them from beck
				--v;
				// store backward index
				vr->component = stackTop;
				// all variants seem to be available
				vr->active = true;
				// and are not blocked
				vr->blocking = 0;
				// initialize physical variant number
				vr->number = v;
				// calculate variant size
				m_list[stackDepth].setVariantNo(v);
				m_list[stackDepth].getSize(vr->xMin, vr->xMax, vr->yMin, vr->yMax);
				vr->size = 
					(((xMax > vr->xMax) ? xMax : vr->xMax) - ((xMin < vr->xMin) ? xMin : vr->xMin) + 1)
					* (((yMax > vr->yMax) ? yMax : vr->yMax) - ((yMin < vr->yMin) ? yMin : vr->yMin) + 1);
				// go for next variant
				++vr;
			} while (v != 0);
		}

		// now sort the variants so that the smallest one is last
		// (because we'll check variants from last to first)
		// there are not quite many variants - nobody should be offended by bubblesort here ;-)
		// also, this is only done once for each comopnent during the check.

		// hey - sort only if we're going to optimize!
		if (optimize)
		{
			VariantRecord * vx = lastVariant - 1;
			VariantRecord * ve = vx;
			while (vx != stackTop->firstVariant)
			{
				if (vx->size <= (vx - 1)->size)
				{
					--vx;
				}
				else
				{
					VariantRecord r;
					memcpy(&r, vx, sizeof(VariantRecord));
					memcpy(vx, (vx - 1), sizeof(VariantRecord));
					memcpy((vx - 1), &r, sizeof(VariantRecord));
					if (vx == ve)
					{
						--vx;
					}
					else
					{
						++vx;
					}
				}
			}
		}

		// after sorting, fill the matrixRow and Index fields
		{
			VariantRecord * vx = lastVariant;
			long v = numVariants;
			do
			{
				--v;
				--vx;
				vx->index = v;
				vx->matrixRow = &m_variantMatrix[v * m_matrixSize];
			} while (vx != stackTop->firstVariant);
		}

		// now:
		// for each variant of the new component
		// check it against each available variant of each existing component
		// and put the result to the matrix

		// of course, this can be done only if this is not the first component
		if (stackDepth == 0)
		{
			// first component
			// nothing can be blocked
			// but we must initialize the free variants count
			stackTop->freeVariants = stackTop->activeVariants;
		}
		else
		{
			{
				ShiftedComponent * sci = &m_list[stackDepth];
				VariantRecord * vx = m_variantVector;
				do
				{
					if (vx->active)
					{
						long c = vx->component->number;
						ShiftedComponent * scc = &m_list[c];

						scc->setVariantNo(vx->number);

						// now go through all variants of the new component
						VariantRecord * vy = stackTop->firstVariant;
						do
						{
							sci->setVariantNo(vy->number);
							// check the two components
							bool ok;
							if ((scc->getDx() == sci->getSx())
								&& (scc->getDy() == sci->getSy())
								&& (scc->getDirection() == sci->getSDirection()))
							{
								// scc precedes sci
								ok = scc->getBasicComponent()->getRegister()->canFollow(
									scc->getVariant(),
									sci->getVariant());
#ifdef CHECKDEBUG
								{
									char buf[80];
									sprintf(buf, "Check: Component %d variant %s vs Component %d variant %s = %s\n",
										vx->component->number,
										scc->getVariant()->getName(),
										vy->component->number,
										sci->getVariant()->getName(),
										ok ? "OK" : "Failed");
									Writer::print(buf);
								}
#endif
							}
							else if ((sci->getDx() == scc->getSx())
								&& (sci->getDy() == scc->getSy())
								&& (sci->getDirection() == scc->getSDirection()))
							{
								// sci precedes scc
								ok = sci->getBasicComponent()->getRegister()->canFollow(
									sci->getVariant(),
									scc->getVariant());
#ifdef CHECKDEBUG
								{
									char buf[80];
									sprintf(buf, "Check: Component %d variant %s vs Component %d variant %s = %s\n",
										vx->component->number,
										scc->getVariant()->getName(),
										vy->component->number,
										sci->getVariant()->getName(),
										ok ? "OK" : "Failed");
									Writer::print(buf);
								}
#endif
							}
							else
							{
								// scc and sci are far from each other - do a bitmap match
								ok = sci->variantCheck(scc);
							}

							// and update the matrix
							vy->matrixRow[vx->index] = ok;
							vx->matrixRow[vy->index] = ok;

							++vy;
						} while (vy != lastVariant);
					}
					++vx;
				} while (vx != stackTop->firstVariant);
			}

			// m_variantMatrix is updated
			// but this might have disabled some variants
			// both in the 'new' and 'old' components
			// now check the matrix:
			// each variant in the new component
			// should have at least one variant available
			// in each 'older' component
			// if some doesn't, disable it
			// also, each 'old' variant
			// should have at least one 'new' variant available

			// first check each 'new' variants

			{
				VariantRecord * jx = stackTop->firstVariant;
				do
				{
					ComponentRecord * cx = m_componentVector;
					bool ok;

					do 
					{
						VariantRecord * vx = cx->firstVariant;

						do
						{
							if (vx->active)
							{
								ok = jx->matrixRow[vx->index];
							}
							else
							{
								// ignore the matrix if the variant isn't active
								ok = false;
							}
							++vx;
						} while (!ok && (vx != cx->lastVariant));
						// ok = true means: there is a working variant in this component
						++cx;
					} while (ok && (cx != stackTop));

					// ok = false means: all the 'c' component variants collide with this variant
					if (!ok)
					{
						jx->active = false;
						stackTop->activeVariants--;
#ifdef CHECKDEBUG
						{
							char buf[80];
							sprintf(buf, "A/ Component %d: eliminating variant %s\n", stackTop->number, m_list[jx->component->number].getVariant(jx->number)->getName());
							Writer::print(buf);
						}
#endif
					}

					// go to next variant
					++jx;
				} while (jx != lastVariant);
			}

			// check if there is at least one variant left
			if (stackTop->activeVariants == 0)
			{
				// no usable variant
				return m_numComponents - maxDepth;
			}

			// now check all 'old' variants
			{
				VariantRecord * vx = stackTop->firstVariant;
				ComponentRecord * cx = stackTop;
				do
				{
					--cx;
					do
					{
						--vx;

						// variant vx of component cx

						// skip disabled variants
						if (vx->active)
						{
							// check all 'new' variants
							VariantRecord * jx = stackTop->firstVariant;
							bool * mx = &vx->matrixRow[jx->index];
							bool ok;

							do 
							{
								if (jx->active)
								{
									ok = *mx;;
								}
								else
								{
									ok = false;
								}
								++mx;
								++jx;
							} while (!ok && (jx != lastVariant));
							// OK=true means: this variant (vx) has at least one matching variant in the new component

							if (!ok)
							{
								// this variant doesn't seem to fit with the new component
								// let's remove it
								vx->active = false;
								--cx->activeVariants;
								if (vx->blocking == 0)
								{
									// remove it from free variants only if it isn't blocked
									// because in such a case it is already removed
									--cx->freeVariants;
								}
#ifdef CHECKDEBUG
								{
									char buf[80];
									sprintf(buf, "B/ Component %d: eliminating variant %s\n", vx->component->number, m_list[vx->component->number].getVariant(vx->number)->getName());
									Writer::print(buf);
								}
#endif
							}
						}
					} while (vx != cx->firstVariant);

					if (cx->activeVariants == 0)
					{
						// there are no variants active for this component
						// no need to continue
						return m_numComponents - maxDepth;
					}
				} while (cx != m_componentVector);
			}

			// update the blocking for the new variants
			// and freeVariants count for the new component

			{
				stackTop->freeVariants = stackTop->activeVariants;

				VariantRecord * jx = stackTop->firstVariant;
				do
				{
					// only for available variants
					if (jx->active)
					{
						// first expect that the variant isn't blocked
						jx->blocking = 0;

						// then go through all old components
						ComponentRecord * cx = m_componentVector;;
						do
						{
							// and check if the component's current variant
							// doesn't block the new variant
							if (!jx->matrixRow[cx->activeVariant->index])
							{
								// if it's the first block on this variant
								if (jx->blocking == 0)
								{
									// decrease number of free variants
									--stackTop->freeVariants;
								}
								// then increase the number ob blocks
								++jx->blocking;
							}
							++cx;
						} while (cx != stackTop);
					}
					++jx;
				} while (jx != lastVariant);
			}

			// now check:
			// - if the current stack contains a variant that turned disabled,
			// backout to that variant
			// - if the new component doesn't have unblocked variants,
			// backout until it has some

			// first find how far we have to backout due to disabled active variant
			{
				// cursor
				ComponentRecord * cx = stackTop;
				// backout marker
				ComponentRecord * bx = stackTop;

				do
				{
					--cx;
					if (!cx->activeVariant->active)
					{
						bx = cx;
					}
				} while (cx != m_componentVector);

				// store the new component
				cx = stackTop;

				// and do the backouting

				while ((stackTop > bx) || (cx->freeVariants == 0))
				{
					// decrease the stack
					--stackTop;
					--stackDepth;
					// backing out:
					// for active variant of the component on the stack top
					// for all variants of all components still not on stack
					// if that

					// index of the checked variant
					VariantRecord * vx = stackTop->firstVariant + stackTop->numVariants;

					if (vx != lastVariant)
					{
						bool * mx = &stackTop->activeVariant->matrixRow[vx->index];
						do
						{
							if (!*mx)
							{
								--vx->blocking;
								if (vx->blocking == 0)
								{
									++vx->component->freeVariants;
								}
							}
							++mx;
							++vx;
						} while (vx != lastVariant);
					}
				}					
			}
		}
	}

NextVariant:
	// invariant:
	// <stackDepth> components are loaded
	// and have a working sequence of variants on the stack
	// <stackDepth+1>th component is loaded
	// and either it has no variant selected
	// or the variant has been scanned

	// find another variant
	{
		VariantRecord * vx = stackTop->activeVariant;
		do
		{
			if (vx == stackTop->firstVariant)
			{
				goto BackOneLevel;
			}
			--vx;
		} while (!vx->active || (vx->blocking != 0));
		stackTop->activeVariant = vx;
	}

	// new variant selected
	// now calculate blocking

	{
		bool ok = true;

		VariantRecord * vx = stackTop->firstVariant + stackTop->numVariants;

		if (vx != lastVariant)
		{
			bool * mx = &stackTop->activeVariant->matrixRow[vx->index];
			do
			{
				if (!*mx)
				{
					if (vx->blocking == 0)
					{
						if (--vx->component->freeVariants == 0)
						{
							ok = false;
						}
					}
					++vx->blocking;
				}
				++mx;
				++vx;
			} while (vx != lastVariant);
		}

		// check if all following components have at least one free variant
		if (ok)
		{
#ifdef CHECKDEBUG
			{
				char buf[80];
				sprintf(buf, "C/ Component %d: selected variant %s\n", stackTop->number, m_list[stackTop->number].getVariant(stackTop->activeVariant->number)->getName());
				Writer::print(buf);
			}
#endif
			// YES
			// we have selected a working variant
			// let's go one level deeper
			if (stackDepth == maxDepth)
			{
				++maxDepth;
			}
			++stackDepth;
			++stackTop;
			goto NextComponent;
		}
	}

// the following code removes all blocks made by the current variant
// allowing to try next variant
UnblockVariant:

	{
		VariantRecord * vx = stackTop->firstVariant + stackTop->numVariants;

		if (vx != lastVariant)
		{
			bool * mx = &stackTop->activeVariant->matrixRow[vx->index];
			do
			{
				if (!*mx)
				{
					--vx->blocking;
					if (vx->blocking == 0)
					{
						++vx->component->freeVariants;
					}
				}
				++mx;
				++vx;
			} while (vx != lastVariant);
		}
	}

	goto NextVariant;

// the following code is executed when the current stack level's possibilities
// have been depleted
// the active variant must not block any variants (must be 'unblocked')

BackOneLevel:

	if (stackDepth == 0)
	{
		// nowhere to backout

		// as long as optimizations are handled elsewhere
		// we clearly didn't find a solution

		// so we van just return

		return m_numComponents - maxDepth;
	}

	// decrease the stack level
	--stackDepth;
	--stackTop;

	goto UnblockVariant;

StartOptimize:

	// now we know that a solution exists
	// all component variants are loaded
	// and matrix is built
	// however, the size optimization needs a completely different algorithm

	// calculate and store the found combination size
	m_area = -1; // force calculation
	minArea = getArea();

	{
		long c = m_numComponents;
		do
		{
			--c;
			m_list[c].storeVariant();
		} while (c != 0);
	}


	// Calculate new minimum size of the result, using only active components
	{
		ComponentRecord * cx = m_componentVector;
		do
		{
			// there is at least one enabled variant in each component
			// so we don't need sentinel to find the first one
			VariantRecord * vx = cx->lastVariant;
			do
			{
				--vx;
			} while (!vx->active);

			// get the variant's size
			long x1 = vx->xMin;
			long x2 = vx->xMax;
			long y1 = vx->yMin;
			long y2 = vx->yMax;

			if (vx != cx->firstVariant)
			{
				do
				{
					--vx;
					if (vx->active)
					{
						if (vx->xMin > x1)
						{
							x1 = vx->xMin;
						}
						if (vx->xMax < x2)
						{
							x2 = vx->xMax;
						}
						if (vx->yMin > y1)
						{
							y1 = vx->yMin;
						}
						if (vx->yMax < y2)
						{
							y2 = vx->yMax;
						}
					}
				} while (vx != cx->firstVariant);
			}

			// x1..y2 contains the component's minimum size
			// if only active variants are considered
			if (x1 < xMin)
			{
				xMin = x1;
			}
			if (x2 > xMax)
			{
				xMax = x2;
			}
			if (y1 < yMin)
			{
				yMin = y1;
			}
			if (y2 > yMax)
			{
				yMax = y2;
			}
			// go for next component if there is one
			++cx;
		} while (cx != lastComponent);
	}

	// check if we didn't reach this minimum already

	if (minArea <= ((xMax - xMin + 1) * (yMax - yMin + 1)))
	{
		// yes, we did - no more search needed
		// the last variant setting also didn't change
		// and the area was calculated
		// so we can just return
		return 0;
	}

	// recalculate variant sizes according to the new minimum size
	// also subtract base area
	{
		VariantRecord * vx = m_variantVector;
		long s = (xMax - xMin + 1) * (yMax - yMin + 1);
		do
		{
			if (vx->active)
			{
				vx->size =
					(((xMax > vx->xMax) ? xMax : vx->xMax) - ((xMin < vx->xMin) ? xMin : vx->xMin) + 1)
					* (((yMax > vx->yMax) ? yMax : vx->yMax) - ((yMin < vx->yMin) ? yMin : vx->yMin) + 1)
					- s;
			}
			++vx;
		} while (vx != lastVariant);
	}

	// remove all variants with the same matrix line with another variant
	// that is contained in this variant's rectangle
	{
		VariantRecord * vx = m_variantVector;
		do
		{
			if (vx->active)
			{
				VariantRecord * vy = vx + 1;
				if (vy != vx->component->lastVariant)
				{
					do
					{
						if (vy->active)
						{
							// vx and vy are two different active variants
							// of the same component
							// let's compare their matrix rows
							bool * mx = vx->matrixRow;
							bool * my = vy->matrixRow;
							VariantRecord * vz = m_variantVector;
							bool ok = true;

							do
							{
								if (vz->active && (vz->component != vx->component))
								{
									ok = *mx == *my;
								}
								++mx;
								++my;
								++vz;
							} while (ok && (vz != lastVariant));

							// ok=true means that the two variants share the same blocking
							if (ok)
							{
								// first check if 'vx' is contained within 'vy'
								// if yes, remove the 'vy' variant

								// if they are of the same size (or inside the circuit)
								// use variant number instead

								bool equal = (((vx->xMin == vy->xMin) || ((xMin <= vx->xMin) && (xMin <= vy->xMin)))
									&& ((vx->xMax == vy->xMax) || ((xMax >= vx->xMax) && (xMax >= vy->xMax)))
									&& ((vx->yMin == vy->yMin) || ((yMin <= vx->yMin) && (yMin <= vy->yMin)))
									&& ((vx->yMax == vy->yMax) || ((yMax >= vx->yMax) && (yMax >= vy->yMax))));
#ifdef CHECKDEBUG
								{
									if (equal)
									{
										Writer::print("Equal size found\n");
									}
								}
#endif

								if ((equal && (vx->number < vy->number))
									|| (!equal
										&& ((vx->xMin >= vy->xMin) || (vx->xMin >= xMin))
										&& ((vx->xMax <= vy->xMax) || (vx->xMax <= xMax))
										&& ((vx->yMin >= vy->yMin) || (vx->yMin >= yMin))
										&& ((vx->yMax <= vy->yMax) || (vx->yMax <= yMax))))
								{
									// vx is contained inside vy
									// remove vy
									vy->active = false;
									--vy->component->activeVariants;
#ifdef CHECKDEBUG
									{
										char buf[80];
										sprintf(buf, "D/ Component %d: eliminating variant %s in favor of %s\n", vy->component->number, m_list[vy->component->number].getVariant(vy->number)->getName(), m_list[vy->component->number].getVariant(vx->number)->getName());
										Writer::print(buf);
									}
#endif
								}
								else if (((vy->xMin >= vx->xMin) || (vy->xMin >= xMin))
									&& ((vy->xMax <= vx->xMax) || (vy->xMax <= xMax))
									&& ((vy->yMin >= vx->yMin) || (vy->yMin >= yMin))
									&& ((vy->yMax <= vx->yMax) || (vy->yMax <= yMax)))
								{
									// vy is contained inside vx
									// remove vy
									vx->active = false;
									--vx->component->activeVariants;
#ifdef CHECKDEBUG
									{
										char buf[80];
										sprintf(buf, "D/ Component %d: eliminating variant %s in favor of %s\n", vx->component->number, m_list[vx->component->number].getVariant(vx->number)->getName(), m_list[vx->component->number].getVariant(vy->number)->getName());
										Writer::print(buf);
									}
#endif
									// no need to continue with inactive variant
									vy = vx->component->lastVariant - 1;
								}
							}
						}
						++vy;
					} while (vy != vx->component->lastVariant);
				}
			}
			++vx;
		} while (vx != lastVariant);
	}

	// calculate component sizes
	// component size is size of its biggest active variant

	{
		ComponentRecord * cx = m_componentVector;
		do
		{
			cx->size = 0;
			VariantRecord * vx = cx->firstVariant;
			do
			{
				if (vx->active)
				{
					if (cx->size < vx->size)
					{
						cx->size = vx->size;
					}
				}
				++vx;
			} while (vx != cx->lastVariant);
			++cx;
		} while(cx != lastComponent);
	}

	// reorder components:
	// put the components with just one variant first
	// then put components sorted by size, the biggest first

	// If you don't like bubblesort, place your favorite sorting algorithm here
	// but as for me, this is done once per check() call
	// and there are not too big amounts of components

	{
		ComponentRecord * cx = lastComponent - 1;

		while (cx != m_componentVector)
		{
			ComponentRecord * c2 = cx->reorder;
			ComponentRecord * c1 = (cx - 1)->reorder;
			// check if c1 < c2
			// c1 < c2 means:
			// either c1 has one active variant
			// or both have more than one and c1's size is smaller than c2's
			// in such a case the variants shouldn't be swapped
			if ((c1->activeVariants == 1) // 
				|| ((c2->activeVariants != 1)
					&& (c1->size >= c2->size)))
			{
				// don't swap, go on
				--cx;
			}
			else
			{
				// swap
				ComponentRecord * t = cx->reorder;
				cx->reorder = (cx - 1)->reorder;
				(cx - 1)->reorder = t;
				
				t = cx + 1;
				if (t != lastComponent)
				{
					cx = t;
				}
				else
				{
					--cx;
				}
			}
		}
	}

	// clear variants

	{
		VariantRecord * vx = m_variantVector;
		do
		{
			// no need to check for inactive variants here
			vx->blocking = 0;
			++vx;
		} while (vx != lastVariant);
	}

	// clean components

	{
		ComponentRecord * cx = m_componentVector;
		do
		{
			cx->freeVariants = cx->activeVariants;
			++cx;
		} while (cx != lastComponent);
	}

	// all preparation steps are done
	// we can just start the recursive search again
	// this time using the reordering
	// and also calculating overall size

	stackTop = m_componentVector;
	realTop = stackTop->reorder;

OptNextComponent:

	if (stackTop == lastComponent)
	{
		// solution found
#ifdef CHECKDEBUG
		{
			Writer::print("Better solution found\n");
		}
#endif
		// and it is smaller than previous solution
		// store it
		ComponentRecord * cx = stackTop - 1;

		minArea = (cx->xMax - cx->xMin + 1) * (cx->yMax - cx->yMin + 1);

		// store variant combination
		// all components, no need to reorder
		cx = m_componentVector;
		do
		{
			m_list[cx->number].setVariantNo(cx->activeVariant->number);
			m_list[cx->number].storeVariant();
			++cx;
		} while (cx != lastComponent);

		goto OptBackOneLevel;
	}

	// solution not found yet
	// initialize variant
	realTop->activeVariant = realTop->lastVariant;

OptNextVariant:

	// get next (working) variant in the topmost component

	{
		do
		{
			if (realTop->activeVariant == realTop->firstVariant)
			{
				// all variants depleted
				// go up
				goto OptBackOneLevel;
			}
			--realTop->activeVariant;
		} while (!realTop->activeVariant->active || (realTop->activeVariant->blocking != 0));

		// new active variant found
		// calculate size

		if (realTop->size == 0)
		{
			// components with zero size don't affect the result size
			// if there is a problem, we don't need another variant
			if (stackTop == m_componentVector)
			{
				stackTop->xMin = xMin;
				stackTop->xMax = xMax;
				stackTop->yMin = yMin;
				stackTop->yMax = yMax;
			}
			else
			{
				stackTop->xMin = (stackTop-1)->xMin;
				stackTop->xMax = (stackTop-1)->xMax;
				stackTop->yMin = (stackTop-1)->yMin;
				stackTop->yMax = (stackTop-1)->yMax;
			}
			long size = (stackTop->xMax - stackTop->xMin + 1) * (stackTop->yMax - stackTop->yMin + 1);
			if (size >= minArea)
			{
				goto OptBackOneLevel;
			}
		}
		else
		{
			if (stackTop == m_componentVector)
			{
				VariantRecord * vx = realTop->activeVariant;
				stackTop->xMin = xMin < vx->xMin ? xMin : vx->xMin;
				stackTop->xMax = xMax > vx->xMax ? xMax : vx->xMax;
				stackTop->yMin = yMin < vx->yMin ? yMin : vx->yMin;
				stackTop->yMax = yMax > vx->yMax ? yMax : vx->yMax;
			}
			else
			{
				VariantRecord * vx = realTop->activeVariant;
				stackTop->xMin = (stackTop-1)->xMin < vx->xMin ? (stackTop-1)->xMin : vx->xMin;
				stackTop->xMax = (stackTop-1)->xMax > vx->xMax ? (stackTop-1)->xMax : vx->xMax;
				stackTop->yMin = (stackTop-1)->yMin < vx->yMin ? (stackTop-1)->yMin : vx->yMin;
				stackTop->yMax = (stackTop-1)->yMax > vx->yMax ? (stackTop-1)->yMax : vx->yMax;
			}
			// this component affects the result size
			// so if we go over, try next variant
			long size = (stackTop->xMax - stackTop->xMin + 1) * (stackTop->yMax - stackTop->yMin + 1);
			if (size >= minArea)
			{
				goto OptNextVariant;
			}
		}
	}
		// the variant seems to be OK
		// do blocking for all following components

	{
		bool ok = true;
		bool * mx = realTop->activeVariant->matrixRow;
		ComponentRecord * cx = stackTop + 1;
		if (cx != lastComponent)
		{
			do
			{
				ComponentRecord * cxr = cx->reorder;
				VariantRecord * vx = cxr->firstVariant;
				VariantRecord * ve = cxr->lastVariant;
				do
				{
					if (vx->active
						&& !mx[vx->index])
					{
						if (0 == vx->blocking++)
						{
							if (0 == --vx->component->freeVariants)
							{
								ok = false;
							}
						}
					}
					++vx;
				} while (vx != ve);
				++cx;
			} while (cx != lastComponent);
		}

		if (ok)
		{
#ifdef CHECKDEBUG
			{
				char buf[80];
				sprintf(buf, "F/ Stack %d - Component %d: selected variant %s\n", stackTop->number, realTop->number, m_list[realTop->number].getVariant(realTop->activeVariant->number)->getName());
				Writer::print(buf);
			}
#endif
			// selected variant is valid
			// let's go one level deeper
			++stackTop;
			realTop = stackTop->reorder;
			goto OptNextComponent;
		}
	}

OptUnblockVariant:

	// remove the blocks caused by the variant on the stack top
	// and go for another variant

	{
		bool * mx = realTop->activeVariant->matrixRow;
		ComponentRecord * cx = stackTop + 1;
		if (cx != lastComponent)
		{
			do
			{
				ComponentRecord * cxr = cx->reorder;
				VariantRecord * vx = cxr->firstVariant;
				VariantRecord * ve = cxr->lastVariant;
				do
				{
					if (vx->active
						&& !mx[vx->index])
					{
						if (0 == --vx->blocking)
						{
							++vx->component->freeVariants;
						}
					}
					++vx;
				} while (vx != ve);
				++cx;
			} while (cx != lastComponent);
		}
	}

	goto OptNextVariant;

OptBackOneLevel:

	// go back by one level
	// there is nothing active on the stack top

	if (stackTop != m_componentVector)
	{
		--stackTop;
		realTop = stackTop->reorder;
		goto OptUnblockVariant;
	}
	else
	{
		// top of stack can't be backed up
		// we're on base level
		// fall down, restore the best result and return

		long c = m_numComponents;
		do
		{
			--c;
			m_list[c].restoreVariant();
		} while (c != 0);
		// force area calculation next time it is called
		m_area = -1;
		// return success
		return 0;
	}
}

void CompoundComponent::printRle(bool forceOrientation)
{
	if (Writer::rleActive())
	{
		long loX, hiX, loY, hiY;

		getSize(loX, hiX, loY, hiY);

		long x;
		long y;
		long i;
		bool * row;
		long x1, y1, x2, y2;

		if (forceOrientation || ((hiX - loX) >= (hiY - loY)))
		{
			row = new bool[hiX - loX + 1];
			assert(row != NULL);

			for (y = loY; y <= hiY; y++)
			{
				for (x = loX; x <= hiX; x++)
				{
					row[x - loX] = false;
				}
				for (i = 0; i < m_numComponents; i++)
				{
					m_list[i].getSize(x1, x2, y1, y2);
					if ((y1 <= y) && (y2 >= y))
					{
						for (x = x1; x <= x2; x++)
						{
							if (m_list[i].getCell(x, y))
							{
								row[x - loX] = true;
							}
						}
					}
				}
				for (x = loX; x <= hiX; x++)
				{
					Writer::putRle(row[x - loX] ? 'o' : 'b');
				}
				Writer::putRle('$');
			}
		}
		else
		{
			row = new bool[hiY - loY + 1];
			assert(row != NULL);

			for (x = loX; x <= hiX; x++)
			{
				for (y = loY; y <= hiY; y++)
				{
					row[y - loY] = false;
				}
				for (i = 0; i < m_numComponents; i++)
				{
					m_list[i].getSize(x1, x2, y1, y2);
					if ((x1 <= x) && (x2 >= x))
					{
						for (y = y1; y <= y2; y++)
						{
							if (m_list[i].getCell(x, y))
							{
								row[y - loY] = true;
							}
						}
					}
				}
				for (y = loY; y <= hiY; y++)
				{
					Writer::putRle(row[y - loY] ? 'o' : 'b');
				}
				Writer::putRle('$');
			}
		}

		delete [] row;


		for (y = 0; y < 50; y++)
		{
			Writer::putRle('$');
		}
		Writer::flushRle();
	}
}

long CompoundComponent::getArea()
{
	if (m_area < 0)
	{
		long xMin, xMax, yMin, yMax;
		getSize(xMin, xMax, yMin, yMax);
		m_area = (xMax - xMin + 1) * (yMax - yMin + 1);
	}
	return m_area;
}

void CompoundComponent::getMinSize(long & xMin, long & xMax, long & yMin, long & yMax)
{
	if (m_numComponents == 0)
	{
		xMin = 0;
		xMax = 0;
		yMin = 0;
		yMax = 0;
	}
	else
	{
		m_list[0].getMinimumSize(xMin, xMax, yMin, yMax);
		long i;
		for (i = 1; i < m_numComponents; i++)
		{
			long x1, x2, y1, y2;
			m_list[i].getMinimumSize(x1, x2, y1, y2);
			if (xMin > x1)
			{
				xMin = x1;
			}
			if (xMax < x2)
			{
				xMax = x2;
			}
			if (yMin > y1)
			{
				yMin = y1;
			}
			if (yMax < y2)
			{
				yMax = y2;
			}
		}
	}
}

long CompoundComponent::getMinArea()
{
	if (m_numComponents == 0)
	{
		return 1;
	}
	else
	{
		long xMin, xMax, yMin, yMax;
		getMinSize(xMin, xMax, yMin, yMax);
		return (xMax - xMin + 1) * (yMax - yMin + 1);
	}
}

void CompoundComponent::getSize(long & xMin, long & xMax, long & yMin, long & yMax)
{
	if (m_numComponents == 0)
	{
		xMin = 0;
		xMax = 0;
		yMin = 0;
		yMax = 0;
	}
	else
	{
		m_list[0].getSize(xMin, xMax, yMin, yMax);
		long i;
		for (i = 1; i < m_numComponents; i++)
		{
			long x1, x2, y1, y2;
			m_list[i].getSize(x1, x2, y1, y2);
			if (xMin > x1)
			{
				xMin = x1;
			}
			if (xMax < x2)
			{
				xMax = x2;
			}
			if (yMin > y1)
			{
				yMin = y1;
			}
			if (yMax < y2)
			{
				yMax = y2;
			}
		}
	}
}

void CompoundComponent::addVariant(BasicComponent * comp, ComponentVariant * var)
{
	addComponent(comp);
	m_list[m_numComponents - 1].setVariant(var);
}

void CompoundComponent::addVariant(long dx, long dy, Direction dir, BasicComponent * comp, ComponentVariant * var)
{
	m_curX = dx;
	m_curY = dy;
	m_curDir = dir;
	addVariant(comp, var);
}


long CompoundComponent::checkMinArea(long maxArea)
{
	if (m_numComponents != 0)
	{
		long xMin, xMax, yMin, yMax;
		m_list[0].getMinimumSize(xMin, xMax, yMin, yMax);

		if (((xMax - xMin + 1) * (yMax - yMin + 1)) > maxArea)
		{
			return m_numComponents;
		}
		
		for (long i = 1; i < m_numComponents; i++)
		{
			long x1, x2, y1, y2;
			m_list[i].getMinimumSize(x1, x2, y1, y2);
			if (xMin > x1)
			{
				xMin = x1;
			}
			if (xMax < x2)
			{
				xMax = x2;
			}
			if (yMin > y1)
			{
				yMin = y1;
			}
			if (yMax < y2)
			{
				yMax = y2;
			}
			if (((xMax - xMin + 1) * (yMax - yMin + 1)) > maxArea)
			{
				return m_numComponents - i;
			}
		}
	}
	return 0;
}
